import json

import lxml.html
import requests

from neolib.http.HTMLForm import HTMLForm


class Page:
    """Represents a HTML web page

    This class is used to encapsulate a HTML web page into a Python object. Not
    all of the attributes a HTML web page can have are present in this class.
    The primary purpose of this class is for creating an encapsulation that
    can be passed around the library for uniformity. The secondary purpose is
    to tie in the requests and lxml libraries into one easy-to-use object.

    Attributes:
        | **url**: The originating URL for this web page

        | **request**: A :class:`Request` object representing the HTTP request
        | **response**: A :class:`Response` object representing the HTTP response

        | **headers**: A dictionary containing the response headers
        | **content**: The HTML content of the web page
        | **json**: If the returned content is JSON this will be loaded

        | **post_data**: Optional dictionary containing any POST data sent
        | **header_values**: Optional dictionary with any header values overriden

        | **usr**: Optional :class:`.User` object used to request this page

        | **document**: An instance of the root element of this HTML page
        | **forms**: A list of :class:`.HTMLForm` objects for forms on this page
    """

    url = ''

    request = None
    response = None

    headers = None
    content = ''
    json = {}

    post_data = {}
    header_values = {}

    usr = None

    document = None
    forms = []

    def __init__(self, url, usr=None, post_data=None,
                 header_values=None, proxy=None):
        """Initializes the page with the given request details

        Args:
            | **url**: The URL of the page to request
            | **usr**: Optional :class:`.User` object to make the request with
            | **post_data**: Optional dictionary of data to POST with (name -> value)
            | **header_values**: Optional dictionary of header values to override the
                request headers with.
            | **proxy**: Optional proxy to use with the request
        """
        # Set class attributes
        self.url = url
        self.post_data = post_data
        self.header_values = header_values

        # Determine if this request is using an existing user
        if usr:
            if post_data:
                r = usr.session.post(url, data=post_data,
                                     headers=header_values, proxies=proxy)
            else:
                r = usr.session.get(url, headers=header_values, proxies=proxy)
        else:
            if post_data:
                r = requests.post(url, data=post_data, headers=header_values,
                                  proxies=proxy)
            else:
                r = requests.get(url, headers=header_values, proxies=proxy)

        # Setup attributes based on response
        self.request = r.request
        self.response = r
        self.headers = r.headers

        # Check what this is
        if r.headers['Content-Type'] == 'application/ajax':
            self.content = r.text
            self.json = json.loads(r.text)
        elif 'text/html' in r.headers['Content-Type']:
            # Prep the html parser
            self.content = r.text
            self.document = lxml.html.document_fromstring(self.content)

            # Process forms
            self.forms = []
            for form in self.xpath('//form'):
                self.forms.append(HTMLForm(url, form))
        else:
            self.content = r.content

    def form(self, **kwargs):
        """Searches for forms this page holds using the given keyword args

        Args:
            **kwargs: The keyword arguments to search for

        Returns:
            A list of :class:`.HTMLForm` instances that match the original query

        Example:
            `form = pg.form(action='/login.phtml')`
        """
        matches = []
        # Loop through all forms and associated attributes looking for
        # supplied keyword arguments
        for form in self.forms:
            for key in kwargs.keys():
                try:
                    value = getattr(form, key)
                    if kwargs[key] == value:
                        matches.append(form)
                except Exception:
                    continue
        return matches

    def xpath(self, query):
        """Queries the current page using the given XPath query

        Args:
            **query**: The XPath query to use

        Returns:
            A list of elements resulting from the XPath query
        """
        return self.document.xpath(query)

    def __repr__(self):
        return "Page <" + self.url + ">"



from neolib.Exceptions import ParseException
from neolib.inventory.Inventory import Inventory
from neolib.item.InventoryItem import InventoryItem
from neolib.item.InventoryItemList import InventoryItemList


class UserInventory(Inventory):
    """ Represents a user's inventory

    Provides an interface for interacting with a user's inventory. Upon calling
    the `load()` method the class will retrieve the contents of the user's
    inventory (excluding NC items) and provide them in a format that is
    consistent with the :class:`Inventory` base class. This class uses the
    :class:`InventoryItem` class for storing item information.
    """

    _log_name = 'neolib.inventory.UserInventory'

    _urls = {
        'inventory': 'http://www.neopets.com/inventory.phtml',
    }

    _paths = {
        'main_inventory': '//table[@class="inventory"][1]//td',
        'item': {
            'id': 'substring-after(./a/@onclick, "(")',
            'img': './a/img/@src',
            'desc': './a/img/@title',
            'name': './text()',
            'rarity': './span/span/text()'
        }
    }

    def __init__(self, usr):
        """ Initializes the parent class """
        super().__init__(usr)

    def load(self):
        """ Loads the inventory for the :class:`User` instance of this class """
        # Load the inventory
        pg = self._get_page('inventory')

        # Loops through all non-NC items
        self.data = []
        try:
            for td in self._xpath('main_inventory', pg):
                id = str(self._xpath('item/id', td)).replace(');', '')
                item = InventoryItem(id, self._usr)

                item.img = str(self._xpath('item/img', td)[0])
                item.desc = str(self._xpath('item/desc', td)[0])
                item.name = str(self._xpath('item/name', td)[0])

                if len(self._xpath('item/rarity', td)) > 0:
                    item.rarity = str(self._xpath('item/rarity', td)[0])
                    item.rarity = item.rarity.replace('(', '').replace(')', '')

                self.data.append(item)
        except Exception:
            self._logger.exception('Unable to parse user\'s inventory', {'pg': pg})
            raise ParseException('Unable to parse user\'s inventory')

    def find(self, **kwargs):
        """ Overrides the :class:`Inventory`:`find()` function to return an
        instance of :class:`InventoryItemList`.

        See the base class's function for more details
        """
        result = super().find(**kwargs)
        return InventoryItemList(self._usr, result)



from neolib.item.Item import Item


class USBackItem(Item):
    """ Represents an item in the user's shop

    Attributes:
        | **stock**: The number of this item stocked in the shop
        | **old_price**: The original price of this item
        | **pos**: The position of this item on the inventory page
        | **pg**: The page number this item is on
        | **remove**: Whether this item has been marked for removal or not
    """

    stock = 0
    old_price = 0
    pos = ''

    pg = 0
    remove = False

    _log_name = 'neolib.item.USBackItem'

    def __init__(self, id, usr, name=""):
        super().__init__(id, usr, name)



from neolib.Exceptions import ParseException
from neolib.inventory.Inventory import Inventory
from neolib.item.USBackItem import USBackItem
from neolib.item.USBackItemList import USBackItemList


class USBackInventory(Inventory):
    """ Represents the inventory of the backend of a user's shop

    Attributes:
        **pages**: The number of pages in the user's inventory
    """

    pages = 0

    _items_per_page = 30

    _log_name = 'neolib.inventory.USBackInventory'

    _urls = {
        'index': 'http://www.neopets.com/market.phtml?type=your',
        'page': 'http://www.neopets.com/market.phtml?order_by=id&type=your&lim=%s',
    }

    _paths = {
        'pages': '//*[@id="content"]/table/tr/td[2]/p[3]/a',
        'inventory': '//form[@action="process_market.phtml"]/table',
        'rows': './tr',
        'details': './td',
        'id': './input[1]/@value',
        'old_price': './input[2]/@value',
        'pos': './input[1]/@name',
    }

    def __init__(self, usr):
        super().__init__(usr)

    def load(self, index=None):
        """ Loads the shop inventory for the :class:`User` instance of this class

        Arguments:
            **index**: Optional :class:`Page` instance with the user's shop
                index page loaded
        """
        # Load the main index
        if not index:
            pg = self._get_page('index')
        else:
            pg = index

        # Reset items if already loaded
        self.data = []

        # Determine how many pages there are.
        pages = len(self._xpath('pages', pg))

        # The first link is not a page
        self.pages = pages - 1

        try:
            # Load the items from the first page
            self._parse_page(pg, 1)

            # Grab from the remaining pages. By starting at 2 we ignore the index.
            for i in range(2, pages):
                pg = self._get_page('page', str(i * self._items_per_page))
                self._parse_page(pg, i)
        except Exception:
            self._logger.exception('Failed to parse user shop', {'pg': pg})
            raise ParseException('Failed to parse user shop')

    def find(self, **kwargs):
        """ Overrides the :class:`Inventory`:`find()` function to return an
        instance of :class:`USBackItemList`.

        See the base class's function for more details
        """
        result = super().find(**kwargs)
        return USBackItemList(self._usr, result)

    def _parse_page(self, pg, pg_num):
        # Grab the main inventory table
        table = self._xpath('inventory', pg)[0]

        # The first and last rows don't contain items
        rows = self._xpath('rows', table)
        rows.pop(0)
        rows.pop()

        # If there's a Pin entry it needs to be removed
        if 'http://images.neopets.com/pin/bank_pin_mgr_35.jpg' in pg.content:
            rows.pop()

        # Grab all items
        for row in rows:
            id = self._xpath('id', row)[0]

            item = USBackItem(id, self._usr)
            details = self._xpath('details', row)

            item.name = str(details[0].xpath('./b/text()')[0])
            item.img = details[1].xpath('./img/@src')[0]
            item.stock = int(details[2].xpath('./b/text()')[0])
            item.type = str(details[3].xpath('./b/text()')[0])
            item.price = int(details[4].xpath('./input/@value')[0])
            item.desc = str(details[5].xpath('./i/text()')[0])

            item.old_price = int(self._xpath('old_price', row)[0])
            item.pos = self._xpath('pos', row)[0].replace('obj_id_', '')
            item.pg = pg_num

            self.data.append(item)



from neolib.item.Item import Item
from neolib.shop.UserFrontShop import UserFrontShop


class WizardItem(Item):
    """ Represents an item from a Shop Wizard search

    Attributes:
        | **owner**: The owner of the item
        | **stock**: The quantity of this item the owner has
    """
    owner = ''
    stock = 0

    def buy(self):
        """ Attempts to buy the item from the owner
        
        Returns:
            Boolean value indicating if the purchase was successful
        """
        # Load the user shop
        us = UserFrontShop(self._usr, self.owner, self.id, self.price)

        # It should be the first item in the list
        if us.inventory[0].name == self.name:
            return us.inventory[0].buy()
        else:
            return False

    def __repr__(self):
        return 'Shop Wizard Result <' + str(self.price) + ">"



from neolib.Exceptions import ParseException
from neolib.inventory.Inventory import Inventory
from neolib.item.SDBItem import SDBItem
from neolib.item.SDBItemList import SDBItemList


class SDBInventory(Inventory):
    """ Represents a user's Safety Deposit Box inventory """
    pages = 0

    _items_per_page = 30

    _log_name = 'neolib.inventory.SDBInventory'

    _urls = {
        'index': 'http://www.neopets.com/safetydeposit.phtml',
        'pages': 'http://www.neopets.com/safetydeposit.phtml?offset=%s&obj_name=&category=0',
        'update': 'http://www.neopets.com/process_safetydeposit.phtml?checksub=scan',
    }

    _paths = {
        'rows': '//*[@id="content"]/table/tr/td[2]/form//tr',
        'details': './td',
        'pages': '//select[@name="offset"][1]/option',
    }

    def load(self):
        # Get the index
        pg = self._get_page('index')

        # Find out how many pages there are
        self.pages = len(self._xpath('pages', pg))

        # Clear any data
        self.data = []

        try:
            # Grab the first page
            self._parse_page(pg, 1)

            # And the rest
            for i in range(1, self.pages):
                pg = self._get_page('pages', str(i * self._items_per_page))
                self._parse_page(pg, i + 1)
        except Exception:
            self._logger.exception('Failed to load SDB inventory', {'pg': pg})
            raise ParseException('Failed to load SDB inventory')

    def update(self):
        # Check for changes
        changed_pages = []
        for i in range(1, self.pages + 1):
            if self._page_change(self.find(pg=i)):
                changed_pages.append(i)

        # Update pages that changed
        if len(changed_pages) > 0:
            for pg_num in changed_pages:
                items = self.find(pg=pg_num)
                data = self._build_data(items, pg_num)

                self._get_page('update', post_data=data)

    def find(self, **kwargs):
        """ Overrides the :class:`Inventory`:`find()` function to return an
        instance of :class:`USBackItemList`.

        See the base class's function for more details
        """
        result = super().find(**kwargs)
        return SDBItemList(self._usr, result)

    def _page_change(self, items):
        for item in items:
            if item.remove:
                return True

        return False

    def _parse_page(self, pg, pg_num):
        # The SDB inventory has some really bad HTML, so this gets ugly
        rows = self._xpath('rows', pg)
        rows = rows[3:-4]

        for row in rows:
            details = self._xpath('details', row)
            id = details[5].xpath('./input/@name')[0].split('[')[1].replace(']', '')

            item = SDBItem(id, self._usr)
            item.img = details[0].xpath('./img/@src')[0]
            item.name = str(details[1].xpath('./b/text()')[0])
            item.desc = str(details[2].xpath('./i/text()')[0])
            item.type = str(details[3].xpath('./b/text()')[0])
            item.stock = int(details[4].xpath('./b/text()')[0])
            item.pg = pg_num

            self.data.append(item)

    def _build_data(self, items, pg_num):
        data = {}

        data['obj_name'] = ''
        data['category'] = '0'
        data['offset'] = str((pg_num - 1) * self._items_per_page)

        for item in items:
            data['back_to_inv[' + item.id + ']'] = str(int(item.remove))

        return data

    def __repr__(self):
        return 'SDB Inventory <' + str(len(self.data)) + ' Items>'



from neolib.Exceptions import ParseException, WizardBanned
from neolib.item.WizardItem import WizardItem
from neolib.item.WizardItemList import WizardItemList
from neolib.NeolibBase import NeolibBase


class Wizard(NeolibBase):
    """ Provides an interface to the Shop Wizard for performing searches """

    SHOP = 'shop'
    GALLERY = 'gallery'

    CONTAINING = 'containing'
    IDENTICAL = 'exact'

    _log_name = 'neolib.shop.Wizard'

    _urls = {
        'index': 'http://www.neopets.com/market.phtml?type=wizard',
    }

    _paths = {
        'rows': '//*[@id="content"]/table/tr/td[2]/div[2]/table[2]/tr',
        'details': './td'

    }

    def __init__(self, usr):
        super().__init__(usr)

    def search(self, item, area='shop', criteria='exact', min=0, max=99999):
        """ Searches the shop wizard with the provided details

        Arguments:
            | **item**: The name of the item to search for
            | **area**: Optional area to search for the item
            | **criteria**: Method to search with
            | **min**: The minimum price of the item
            | **max**: The maximum price of the item

        Returns:
            Instance of :class:`WizardItemList` with the search results or None
                if the search returned zero results
        """
        # Load the index
        pg = self._get_page('index')

        # Grab the form and set the values
        form = pg.form(action='market.phtml')[0]
        data = {
            'type': 'process_wizard',
            'feedset': '0',
            'shopwizard': item,
            'table': area,
            'criteria': criteria,
            'min_price': min,
            'max_price': max,
        }

        form.update(data)

        # Submit the form and get the results
        pg = form.submit(self._usr)

        # Check for results
        if 'I did not find anything' in pg.content:
            return

        # Check for ban
        if 'too many searches' in pg.content:
            raise WizardBanned

        # Parse the results
        try:
            rows = self._xpath('rows', pg)
            rows.pop(0)

            items = []
            for row in rows:
                details = self._xpath('details', row)

                url = details[0].xpath('./a/@href')[0]
                id = url.split('obj_info_id=')[1].split('&')[0]

                item = WizardItem(id, self._usr)
                item.owner = str(details[0].text_content())
                item.name = str(details[1].text_content())
                item.stock = int(details[2].text_content())
                item.price = int(self._remove_multi(details[3].text_content(), [',', ' NP']))

                items.append(item)
        except Exception:
            self._logger.log('Failed to parse shop wizard results', {'pg': pg})
            raise ParseException('Could not parse shop wizard results')

        return WizardItemList(self._usr, items)

    def __repr__(self):
        return "Shop Wizard"



from neolib.user.hooks.Hook import Hook
from neolib.user.Neopet import Neopet


class UserDetails(Hook):
    """A hook for grabbing the user's neopoints and active pet"""

    _log_name = 'neolib.user.hooks.UserDetails'

    _paths = {
        'nps': '//*[@id="npanchor"]',
        'neopet_name': '//*[@id="content"]/table/tr/td[1]/div[1]/table/tr[1]/td/a/b',
        'neopet_info': '//table[@class="sidebarTable"][1]',
    }

    _regex = {
        'neopet': {
            'species': 'Species:.*?<b>(.*?)</b>',
            'health': 'Health:.*?<font.*?<b>(.*?)</b>',
            'mood': 'Mood:.*?<b>(.*?)</b>',
            'hunger': 'Hunger:.*?<b>.*?">(.*?)<',
            'age': 'Age:.*?<b>(.*?) days</b>',
            'level': 'Level:.*?<b>(.*?)</b>',
        }
    }

    def __init__(self):
        super().__init__()

    def execute(self, usr, pg):
        """A hook for grabbing the user's neopoints and active pet"""
        # First get the neopoints if available
        try:
            if "NP: <a id='npanchor'" in pg.content:
                nps = self._xpath('nps', pg)[0].text
                usr.neopoints = int(nps.replace(',', ''))
        except Exception:
            self._logger.exception('Unable to parse user\'s neopoints')
            return  # This is not critical enough to warrant re-raising

        # Next try for the active pet
        try:
            if 'activePetInfo' in pg.content:
                usr.active_pet = Neopet()
                usr.active_pet.name = self._xpath('neopet_name', pg)[0].text

                html = self._path_to_html('neopet_info', pg)
                for key in self._regex['neopet'].keys():
                    if key == 'age' or key == 'level':
                        value = int(self._search('neopet/' + key, html, True)[0].replace(',', ''))
                    else:
                        value = self._search('neopet/' + key, html, True)[0]
                    setattr(usr.active_pet, key, value)
        except Exception:
            self._logger.exception('Unable to parse active pet info')
            return  # This is not critical enough to warrant re-raising



from collections import UserDict
from urllib.parse import urlparse

import lxml


class HTMLForm(UserDict):
    """Represents a HTML form on a web page

    This class is used to encapsulate an HTML form into a Python object. It
    does so by extracting the important form attributes (action, method) as
    well as any form inputs that actually get sent back to the web server. The
    form inputs can be accessed directly by name as an attribute of this
    object (I.E form.input_name)

    Attributes
       | **action**: The destination of the form
       | **method**: The method in which the form is submitted
       | **url**: The base URL this form came from
       | **data**: A dictionary containing all of the inputs of the field. The
              dictionary uses the input name as the key and an instance of
              :class:`.HTMLFormInput` as the value.
    """

    action = ''
    method = ''
    url = ''
    data = {}

    def __init__(self, base_url, form_element):
        """Initializes the form with the given base URL and form element

        Args:
            | **base_url**: The base URL this form originated from
            | **form_element**: An instance of a form element
        """
        # Form setup
        self.url = base_url

        if len(form_element.xpath('./@action')) > 0:
            self.action = form_element.xpath('./@action')[0]
        if len(form_element.xpath('./@method')) > 0:
            self.method = form_element.xpath('./@method')[0]

        # Grab all the inputs
        self.data = {}
        for einp in form_element.xpath('.//input | .//select'):
            # Search for any attributes and assign them as necessary
            inp = HTMLFormInput()
            for attribute in dir(inp):
                if len(einp.xpath('./@' + attribute)) > 0:
                    setattr(inp, attribute, einp.xpath('./@' + attribute)[0])

            # Look for default values on select inputs
            if type(einp) is lxml.html.SelectElement:
                if einp.xpath('./option[@selected]'):
                    inp.value = einp.xpath('./option[@selected]/@value')[0]

            if inp.name:
                self.data[inp.name] = inp

    def update(self, data=None, **kwargs):
        """Updates the stored fields with the given data

        Args:
            **data**: Optional dictionary to use instead of keyword arguments
            **kwargs**: Keyword arguments to use for updating field values
        """
        if data:
            d = data
        else:
            d = kwargs

        for key in d.keys():
            if key not in self.data:
                inp = HTMLFormInput()
                inp.name = key
                inp.value = d[key]
                self.data[key] = inp
            else:
                self.data[key].value = d[key]

    def submit(self, usr):
        """Submits the current form as if the user had pressed the submit button

        This method will utilize the provided :class:`User` instance to POST
        the gathered form data to the appropriate URL as determined by the base
        URL provided and the action of the form itself. It returns the result
        of the POST action.

        Args:
            **usr**: An instance of :class:`User` to use for submitting the form

        Returns:
            A instance of class:`Page` representing the result
        """
        # Assemble the post data
        post_data = {}
        for key in self.data.keys():
            post_data[key] = self.data[key].value

        # Assemble the address to post to
        u = urlparse(self.url)
        if not self.action:
            self.action = self.url
        elif self.action == u.path:
            self.action = self.url
        else:
            if u.netloc not in self.action:
                path = '/'.join(u.path.split('/')[1:-1])
                if self.action.startswith('/'):
                    path = path + self.action
                else:
                    path = path + '/' + self.action
                self.action = 'http://' + u.netloc + '/' + path

        # Return a new page with the result of the form submission
        return usr.get_page(self.action, post_data)

    def __repr__(self):
        return "HTML Form <" + self.action + ">"


class HTMLFormInput:
    """Represents an input element of an HTML form

    Attributes
       | **type**: The input type
       | **name**: The input name
       | **value**: The input value
       | **x**: The x-coordinate if this input is an image
       | **y**: The y-coordinate if this input is an image
    """

    type = ''
    name = ''
    value = ''

    x = 0
    y = 0

    def __repr__(self):
        return "HTML Input <" + self.name + "(" + self.type + ")>"



from collections import UserList

from neolib.NeolibBase import NeolibBase


class ItemList(NeolibBase, UserList):
    """"A base class for returning a list of items

    This class should be inherited when returning a list of items from any
    type of search or query. The goal is to provide specialized functions for
    dealing with items in large quantities. For instance, allowing an end-user
    to buy all items from a shop wizard result with one command to this class.

    Attributes
        | **data**: The list of returned items
    """

    data = []

    def __init__(self, usr, items):
        super().__init__(usr)
        self.data = items



import random

import requests

from neolib.Exceptions import (AccountFrozen, BirthdayLocked, InvalidBirthday,
                               NeopetsOffline, NoActiveNeopet)
from neolib.http.Page import Page
from neolib.inventory.SDBInventory import SDBInventory
from neolib.inventory.UserInventory import UserInventory
from neolib.NeolibBase import NeolibBase
from neolib.shop.UserBackShop import UserBackShop
from neolib.user.Bank import Bank
from neolib.user.hooks.UserDetails import UserDetails
from neolib.user.Profile import Profile


class User(NeolibBase):
    """Represents a Neopets user account

    This is the one class that ties most of the library together. Every object
    that inherits the library's base class has a dedicated attribute for
    holding a User object. This class should be used as the representation of
    the Neopets user account and all actions involving the account should
    reference this class.

    Attributes
       | **username**: The account username
       | **password**: The account password
       | **pin**: The pin number for the account if applicable

       | **session**: The HTTP session for the account

       | **neopoints**: The number of neopoints the user has on hand

       | **active_pet**: :class:`.Neopet` object representing the user's active pet

       | **profile**: :class:`.Profile` object representing the user's account profile
       | **mail**: :class:`Neomail` object for interacting with the user's neomail

       | **inventory**: :class:`UserInventory` object representing the user's inventory
       | **sdb**: :class:`SDB` object representing the user's safety deposit box
       | **shop**: :class:`UserShop` object representing the user's shop
       | **bank**: :class:`Bank` object representing the user's bank

       | **trades**: :class:`TradingPost` object for interacting with the user's trades
       | **auctions**: :class:`AuctionHouse` object for interacting with the user's auctions

       | **hooks**: A list of :class:`Hook` based classes that will be executed
            after requesting a page
    """

    username = ''
    password = ''
    pin = ''

    session = ''

    neopoints = 0

    active_pet = None

    _profile = None
    mail = None

    _inventory = None
    _SDB = None
    _shop = None
    _bank = None

    trades = None
    auctions = None

    hooks = []

    _last_page = ''

    _log_name = 'neolib.user.User'

    _urls = {
        'index': 'http://www.neopets.com/',
    }

    @property
    def profile(self):
        if not self._profile:
            self._profile = Profile(self)
            self._profile.load()

        return self._profile

    @property
    def inventory(self):
        if not self._inventory:
            self._inventory = UserInventory(self)
            self._inventory.load()

        return self._inventory

    @property
    def SDB(self):
        if not self._SDB:
            self._SDB = SDBInventory(self)
            self._SDB.load()

        return self._SDB

    @property
    def shop(self):
        if not self._shop:
            self._shop = UserBackShop(self)
            self._shop.load()

        return self._shop

    @property
    def bank(self):
        if not self._bank:
            self._bank = Bank(self)
            self._bank.load()

        return self._bank

    def __init__(self, username, password='', pin=None):
        """Initializes the user with the given username, password, and pin

        Args:
            | **username**: The username for the account
            | **password**: The password for the account
            | **pin**: Optional pin number for the user's account
        """
        # Initialize parent
        super().__init__()

        # Set username and password
        self.username, self.password = username, password

        # Grab the pin if available
        self.pin = pin if pin else None

        # Initialize session
        self.session = requests.session()

        # Setup default hooks
        self.add_hook(UserDetails)

    def login(self, birthday=None):
        """Performs a login and returns the result

        This function will submit the login form on the Neopets website with
        the user's username and password. It then returns the status of the
        login by checking for the user's username on the resulting page. This
        method must be called before doing any other account activities that
        normally require the user to be logged in.

        Arguments:
            birthday: `Date` object with the user's birthday for accessing a
                locked account

        Returns:
            A boolean value indicating if the login was successful
        """
        # TNT has very tight anti-cheat controls so in this scenario it
        # is best to simulate a legitimate login by navigating to the
        # index page first
        pg = self.get_page(self._urls['index'])

        # Fill in the login form
        form = pg.form(action='/login.phtml')[0]
        form.update(username=self.username, password=self.password)

        # Submit the form
        pg = form.submit(self)

        # Check if the account is frozen
        if 'account has been FROZEN' in pg.content:
            self._logger.error('User account ' + self.username + ' is frozen')
            raise AccountFrozen('Account is frozen')

        # Check for password strength
        if 'STOP!  Your password' in pg.content or 'username matches your e-mail' in pg.content:
            # We're actually logged in now, we can ignore this page and just
            # request the index again
            self._logger.warning('User ' + self.username + ' has an unsecured password')
            pg = self.get_page('http://www.neopets.com/')

        # Check if account is birthday locked and if a birthday was given
        if 'please verify your birthday' in pg.content and not birthday:
            self._logger.error('User ' + self.username + ' is birthday locked')
            raise BirthdayLocked('Account is birthday locked')
        elif 'please verify your birthday' in pg.content:
            self._logger.warning('User ' + self.username + ' is birthday locked. Attempting to submit date')
            pg = self._submit_birthday(pg, birthday)

            if 'Invalid birthday' in pg.content:
                self._logger.error('Birthday failed for user ' + self.username)
                raise InvalidBirthday('Invalid birthday given')

        # Check if we've maxed out birthday tries
        if 'birthdate for this account incorrectly 3 times today' in pg.content:
            self._logger.error('Birthday failed for user ' + self.username)
            self._logger.warning('Birthday locked for one day')
            raise InvalidBirthday('Birthday locked for one day')

        # Check for Lawyerbot
        if 'Lawyerbot message' in pg.content:
            # We have to take one extra step and accept the EULA
            self._logger.warning('Lawyerbot detected. Accepting EULA.')
            form = pg.form(action='/accept_terms.phtml')[0]
            form.update(accept='1')
            pg = form.submit(self)

        # If the account doesn't have a Neopet, we will be redirected to create
        # one on the Neopet creation page
        if pg.response.url == 'http://www.neopets.com/reg/page4.phtml':
            self._logger.error('No active Neopet for user ' + self.username)
            raise NoActiveNeopet('Missing active pet')

        # Return if it was successful
        return pg.response.url == 'http://www.neopets.com/'

    def get_page(self, url, post_data='', header_values=''):
        """A wrapper function that returns a page using the user's session

        This method should be used over initializing a new page object by
        supplying the user's session. It performs checks to inject the user's
        pin number at the appropriate time as well as checks for random events
        and acts on them accordingly.

        Args:
            | **url**: The url of the page to request
            | **post_data**: Optional dictionary containing post data to POST
            | **header_values**: Optional dictionary to override header values

        Returns:
            A :class:`.Page` object representng the requested page
        """
        # Inject a referer (Neopets looks for these often)
        if not header_values and self._last_page:
            header_values = {'Referer': self._last_page}
        elif "Referer" not in header_values and self._last_page:
            header_values['Referer'] = self._last_page

        pg = Page(url, self, post_data=post_data, header_values=header_values)

        # Check if this is an HTML page
        if type(pg.content) is bytes:
            return pg

        self._last_page = url

        # This image is shown if Neopets is offline
        if "http://images.neopets.com/homepage/indexbak_oops_en.png" in pg.content:
            raise NeopetsOffline

        # Call hooks
        for hook in self.hooks:
            h = hook()
            h.execute(self, pg)

        return pg

    def add_hook(self, hook):
        """Adds an instance of a :class:`Hook` based class to be executed after
        page calls
        """
        self.hooks.append(hook)

    def _submit_birthday(self, pg, birthday):
        """ Submits a user's birthday for authentication

        Arguments:
            birthday: Date object with the user's birthday

        Returns:
            Page object with the result of the submission
        """
        # Update the form with birthday values and password
        form = pg.form(action='/login.phtml')[1]
        month = '%02d' % birthday.month
        day = '%02d' % birthday.day

        form.update(password=self.password, dob_m=month,
                    dob_y=str(birthday.year), dob_d=day)

        # Random x,y coordinates
        x = random.randint(10, 30)
        y = random.randint(10, 30)
        form.update(x=str(x), y=str(y))

        return form.submit(self)

    def __repr__(self):
        return "User <" + self.username + ">"




class Neopet:
    """Represents a Neopet

    Attributes:
        | **name**: The neopet's name
        | **gender**: The neopet's gender
        | **species**: The neopet's species
        | **age**: The neopet's age
        | **level**: The neopet's level

        | **health**: The neopet's current health
        | **mood**: The neopet's current mood
        | **hunger**: The neopet's current hunger level
    """
    name = ''
    gender = ''
    species = ''
    age = 0
    level = 0

    health = ''
    mood = ''
    hunger = ''



class Transaction():
    """ Represents a transaction that occurred in a user's sales history

    Attributes:
        **date**: The date the transactions occurred
        **item**: The item that was purchased
        **buyer**: The user who bought the item
        **price**: The price the item was sold at (integer)
    """
    date = ''
    item = ''
    buyer = ''
    price = 0

    def __repr__(self):
        return 'Transaction <' + self.item + '>'



from neolib.NeolibBase import NeolibBase


class Item(NeolibBase):
    """A base class for templating Neopets items

    This class should be inherited when creating any type of Neopet item. While
    items share many attributes, not all pages on Neopets portray items with
    the exact same attributes (for instance the user inventory is the only
    place where weight is shown). This class includes attributes that are found
    on most pages. Classes that inherit this class should add their own unique
    attributes and attempt to fill in all attributes provided here.

    Attributes
        | **id**: The identifying number for the item (may be different
            depending on where this item is located)
        | **name**: The item name
        | **desc**: The item description
        | **price**: The item price (may not be applicable)
        | **img**: The URL to the image of the item
        | **type**: The item type
    """
    id = ''
    name = ''
    desc = ''
    price = 0
    img = ''
    type = ''

    def __init__(self, id, usr, name=""):
        """ Initializes the class

        Initializes the parent class with the given :class:`User` instance
        and the current class with the item ID and name

        Args
            | **id**: The item ID
            | **usr**: The :class:`User` instance that is using this item
            | **name**: Optional item name
        """
        super().__init__(usr)

        self.id = id
        self.name = name

    def __repr__(self):
        return "Item <" + self.name + ">"



from neolib.Exceptions import ParseException
from neolib.inventory.USBackInventory import USBackInventory
from neolib.NeolibBase import NeolibBase
from neolib.shop.History import History


class UserBackShop(NeolibBase):
    """ Represents the backend of a user shop

    This class provides an interface to the backend of a user's shop. This
    includes access to all shop details, shop inventory, and additional actions
    like updating the inventory or changing the shop details.

    Attributs:
        | **name**: The shop name
        | **size**: The shop size
        | **keeper_img**: The image for the shop keeper
        | **keeper_message**: The shop keeper's message
        | **stocked**: The total number of items stocked in the shop
        | **free_space**: The total amount of free space in the shop
        | **created**: If the shop has been created or not (False if user does
            not currently have a shop)
        | **inventory**: Instance of :class:`USBackInventory`
        | **till**: The user's current till (updated everytime it's read)
        | **history**: The sales history for the user's shop (:class:`History`)
        | **upgrade_cost**: The cost to upgrade the shop to the next size
    """
    name = ''
    size = 0
    keeper_img = ''
    keeper_name = ''
    keeper_message = ''
    stocked = 0
    free_space = 0

    created = True

    inventory = None

    _history = None

    _log_name = 'neolib.shop.UserBackShop'

    _urls = {
        'index': 'http://www.neopets.com/market.phtml?type=your',
        'update': 'http://www.neopets.com/process_market.phtml',
        'till': 'http://www.neopets.com/market.phtml?type=till',
        'edit': 'http://www.neopets.com/market.phtml?type=edit',
    }

    _paths = {
        'shop': {
            'name': '//*[@id="content"]/table/tr/td[2]/b/text()',
            'size': '//*[@id="content"]/table/tr/td[2]/text()[9]',
        },
        'keeper': {
            'img': '//img[@name="keeperimage"]/@src',
            'name_msg': '//*[@id="content"]/table/tr/td[2]/center[1]/b[1]/text()',
        },
        'stock': {
            'stocked': '//*[@id="content"]/table/tr/td[2]/center[1]/b[2]/text()',
            'free_space': '//*[@id="content"]/table/tr/td[2]/center[1]/b[3]/text()',
        },
        'upgrade': {
            'cost': '//form[@action="process_market.phtml"]/input[@type="submit"]/@value',
        },
        'till': '//*[@id="content"]/table/tr/td[2]/p[1]/b/text()',
    }

    @property
    def till(self):
        pg = self._get_page('till')

        try:
            return int(self._xpath('till', pg)[0].replace(' NP', ''))
        except Exception:
            self._logger.exception('Failed to parse user till', {'pg': pg})
            raise ParseException('Failed to parse user till')

    @property
    def history(self):
        if not self._history:
            self._history = History(self._usr)
            self._history.load()

        return self._history

    @property
    def upgrade_cost(self):
        pg = self._get_page('edit')

        amt = self._xpath('upgrade/cost', pg)[0].split(': ')[1].replace(' NP', '')
        return int(amt)

    def __init__(self, usr):
        super().__init__(usr)

    def load(self):
        """ Loads the user's shop details and inventory """
        # Load the index
        pg = self._get_page('index')

        # Check if they have a shop
        if 'You don\'t have your own shop yet!' in pg.content:
            # Initialize empty inventory and set status
            self.created = False
            self.inventory = []
            self._logger.warning('User ' + self._usr.username + ' does not have a shop')
            return

        # Load the main details
        self.name = self._xpath('shop/name', pg)[0]
        self.size = int(self._xpath('shop/size', pg)[0].split('size ')[1].replace(')', ''))
        self.keeper_img = self._xpath('keeper/img', pg)[0]
        self.keeper_name = self._xpath('keeper/name_msg', pg)[0].split(' says')[0]
        self.keeper_message = self._xpath('keeper/name_msg', pg)[0].split('says ')[1].replace('\'', '')
        self.inventory = USBackInventory(self._usr)

        # Load the inventory
        self.inventory = USBackInventory(self._usr)

        # Check for stock
        if 'There are no items in your shop!' in pg.content:
            return

        # Load the inventory
        self.stocked = int(self._xpath('stock/stocked', pg)[0])
        self.free_space = int(self._xpath('stock/free_space', pg)[0])

        self.inventory.load(pg)

    def update(self):
        """ Looks for any changes in the user's inventory and updates them

        This method will move through all items per inventory page looking for
        changes in either the delete attribute or a difference between the
        old price and current price of an item (indicating a price change). It
        will then submit only the pages that have items which changed.
        """
        # Determine if any pages have changed
        changed_pages = []
        for i in range(1, self.inventory.pages + 1):
            if self._page_change(self.inventory.find(pg=i)):
                changed_pages.append(i)

        # Update the pages with changes
        if len(changed_pages) > 0:
            for pg_num in changed_pages:
                items = self.inventory.find(pg=pg_num)
                data = self._build_data(items, pg_num)

                self._get_page('update', post_data=data)

    def withdraw(self, amount):
        """ Withdraws the given amount from the user's shop till

        Arguments:
            **amount**: The amount to withdraw

        Returns:
            Boolean indicating if the withdrawal was successful
        """
        # Build the data
        data = {}
        data['type'] = 'withdraw'
        data['amount'] = str(amount)

        # Submit the withdrawal
        pg = self._get_page('update', post_data=data)

        # Return the status
        if 'red_oops.gif' in pg.content:
            return False
        else:
            return True

    def upgrade(self):
        """ Upgrades the user's shop to the next size

        Returns:
            Boolean indicating whether or not the upgrade was successful
        """
        data = {}
        data['type'] = 'upgrade'

        pg = self._get_page('update', post_data=data)

        if 'red_oops.gif' in pg.content:
            return False
        else:
            return True

    def _page_change(self, items):
        for item in items:
            # A difference between price and old_price indicates a price change
            # A change in the remove attribute indicates an item needs removing
            if item.old_price != item.price:
                return True
            elif item.remove:
                return True

        return False

    def _build_data(self, items, pg_num):
        data = {}

        data['type'] = 'update_prices'
        data['order_by'] = 'id'
        data['view'] = ''
        data['limit'] = str(pg_num * self.inventory._items_per_page)

        for item in items:
            data['obj_id_' + item.pos] = item.id
            data['oldcost_' + item.pos] = str(item.old_price)
            data['cost_' + item.pos] = str(item.price)
            data['back_to_inv[' + item.id + ']'] = str(int(item.remove))

        return data

    def __repr__(self):
        return 'User Shop Backend <' + str(len(self.inventory)) + ' items stocked>'



from collections import UserList

from neolib.item.ItemList import ItemList


class USFrontItemList(ItemList):
    """ Represents a list of items returned from querying another user's shop """

    _log_name = 'neolib.item.USFrontItemList'

    def buy(self):
        """ Attempts to buy all items from this list

        Returns:
            List of items that were successfully bought
        """
        success = []
        for item in self.data:
            if item.buy():
                success.append(item)

        return success

    def __getitem__(self, item):
        result = UserList.__getitem__(self, item)
        if type(result) is list:
            if len(result) > 1:
                return USFrontItemList(self._usr, result)
        else:
            return result



from collections import UserList

from neolib.item.ItemList import ItemList


class USBackItemList(ItemList):
    """ Represents a returned list of items from querying a user's shop """

    _log_name = 'neolib.item.USBackItemList'

    def __init__(self, usr, items):
        super().__init__(usr, items)

    def remove(self):
        """ Sets all items in the current list to be removed in the next update """
        for item in self.data:
            item.remove = True

    def __getitem__(self, item):
        result = UserList.__getitem__(self, item)
        if type(result) is list:
            if len(result) > 1:
                return USBackItemList(self._usr, result)
        else:
            return result



import logging
import logging.config
from neolib.Filters import HTMLFilter
import os


def set_config(config):
    logging.config.dictConfig(config)

# The below variables should be imported and modified by developers to configure
# the internal logging. For further information on how this dictionary should
# be configured please refer to the python help documents.
LOG_DIR = 'neolib_logs'
LOG_FILE = 'neolib.log'
PAGE_DIR = 'pages'

CONFIG = {'version': 1,
          'formatters': {
              'default': {
                  'format': '[%(asctime)s] %(name)s (%(levelname)s) Line %(lineno)d in %(filename)s : %(message)s'
                  },
              },
          'filters': {
              'html_filter': {
                  'name': 'neolib',
                  '()': HTMLFilter,
                  },
              },
          'handlers': {
              'console': {
                  'class': 'logging.StreamHandler',
                  'level': 'DEBUG',
                  'formatter': 'default',
                  'filters': ['html_filter'],
                  },
              'file': {
                  'class': 'logging.handlers.TimedRotatingFileHandler',
                  'level': 'DEBUG',
                  'formatter': 'default',
                  'filters': ['html_filter'],
                  'filename': LOG_DIR + '/' + LOG_FILE,
                  'when': 'D',
                  },
              },
          'loggers': {
              'neolib': {
                  'level': 'DEBUG',
                  'handlers': ['console', 'file'],
                  },
              },
          }

# Check for the log directory
if not os.path.isdir(LOG_DIR):
    os.makedirs(LOG_DIR)

# Check for pages directory
if not os.path.isdir(LOG_DIR + '/' + PAGE_DIR):
    os.makedirs(LOG_DIR + '/' + PAGE_DIR)

# Set the configuration for the base logger 'neolib'
set_config(CONFIG)



from neolib.inventory.USFrontInventory import USFrontInventory
from neolib.item.USFrontItem import USFrontItem
from neolib.NeolibBase import NeolibBase


class UserFrontShop(NeolibBase):
    """ Provides an interface for interacting with another user's shop

    Attributes:
        | **inventory**: Instance of :class:`USFrontInventory` with the shop's
            inventory loaded.
    """

    inventory = None

    _log_name = 'neolib.shop.UserFrontShop'

    _urls = {
        'shop': 'http://www.neopets.com/browseshop.phtml?owner=%s',
        'shop_item': 'http://www.neopets.com/browseshop.phtml?owner=%s&buy_obj_info_id=%s&buy_cost_neopoints=%s'
    }

    _paths = {
        'main_item': '//*[@id="content"]/table/tr/td[2]/div[4]/table/tr/td',
    }

    def __init__(self, usr, owner, item_id='', price=''):
        """ Loads the shops inventory and grabs main item if necessary

        The initialization of this class takes multiple parameters for different
        scenarios. If a shop needs to be loaded (including all shop pages) then
        the `usr` and `owner` arguments should only be supplied. If a specific
        item in the shop is being sought after (like from a shop wizard result)
        then the item id and price should also be supplied. In the latter case
        only the first page of the user shop is loaded and if the item is found
        it will be appended to the very beginning of the inventory list.

        Arguments:
            | **usr**: The :class:`User` instance to use
            | **owner**: The username of the owner of the shop
            | **item_id**: The item id to search for
            | **price**: The price of the item to search for
        """
        super().__init__(usr)

        # Load the appropriate page
        if item_id and price:
            pg = self._get_page('shop_item', (owner, item_id, str(price)))
        else:
            pg = self._get_page('shop', owner)

        # Load the inventory
        self.inventory = USFrontInventory(self._usr)

        if item_id and price:
            self.inventory.load(owner, pg)
        else:
            self.inventory.load(owner, pg, True)

        # If searching for a specific item, find it
        if item_id and price:
            try:
                td = self._xpath('main_item', pg)[0]
                item = USFrontItem('', self._usr)

                item.url = td.xpath('./a/@href')[0]
                item.name = str(td.xpath('./b/text()')[0])
                item.stock = int(td.xpath('./text()[3]')[0].replace(' in stock', ''))
                item.price = int(self._remove_multi(td.xpath('./text()[4]')[0], [',', ' NP', 'Cost : ']))

                # Add it to the beginning of the inventory
                self.inventory.data = [item] + self.inventory.data
            except Exception:
                # This most likely means the item has already been bought
                pass



from collections import UserList

from neolib.NeolibBase import NeolibBase


class Inventory(NeolibBase, UserList):
    """A base class for templating how an inventory should operate

    This class should be inherited when parsing and delivering any inventory
    to an end-user. An inventory in this case is a set of items that the
    end-user may interact with (i.e shop inventory, SDB inventory, etc).

    Attributes
        | **data**: A list of items in the inventory
        | **query_types**: The valid queries that can be supplied to find()
    """
    data = []

    query_types = [
        'contains',
        'startswith',
        'endswith'
        'gt',
        'lt',
    ]

    def load(self):
        """ Loads the current inventory

        This function should be overriden by children classes and used to load
        items with the content of the inventory
        """
        pass

    def find(self, **kwargs):
        """ Searches the current inventory using keyword arguments

        Loops through each keyword argument and searches the current inventory
        for any item that has an attribute (key name) that matches a value
        (key value).

        Args
            **kwargs**: Keyword arguments used for searching

        Returns
            A list of items that match the search

        Example
            carrot = usr.inventory.item(name='carrot')[0]
        """
        matches = []

        for item in self.data:
            match = 0
            for key in kwargs.keys():
                try:
                    # Determine if this query has extra arguments
                    if '__' in key:
                        arg = key.split('__')[1]
                        atrib = key.split('__')[0]
                        value = getattr(item, atrib)

                        if arg not in self.query_types:
                            continue

                        if arg == 'contains' and type(value) is str:
                            if kwargs[key].lower() in value.lower():
                                match += 1
                        elif arg == 'startswith' and type(value) is str:
                            if value.lower().startswith(kwargs[key].lower()):
                                match += 1
                        elif arg == 'endswith' and type(value) is str:
                            if value.lower().endswith(kwargs[key].lower()):
                                match += 1
                        elif arg == 'gt' and self._is_int(value):
                            if int(value) > int(kwargs[key]):
                                match += 1
                        elif arg == 'lt' and self._is_init(value):
                            if int(value) < int(kwargs[key]):
                                match += 1
                    else:
                        value = getattr(item, key)
                        if type(kwargs[key]) is str:
                            if kwargs[key].lower() == value.lower():
                                match += 1
                        else:
                            if kwargs[key] == value:
                                match += 1
                except Exception:
                    continue

            if match == len(kwargs.keys()):
                matches.append(item)
        return matches

    def __init__(self, usr):
        super().__init__(usr)

    def __repr__(self):
        return "Inventory <" + str(len(self)) + " items>"



from neolib.Exceptions import (AttributeNotFound, InvalidDetails, InvalidEmail,
                               InvalidNeopet, InvalidPassword,
                               MissingRequiredAttribute, NeopetNotAvailable,
                               UsernameNotAvailable)
from neolib.NeolibBase import NeolibBase
from neolib.user.User import User


class RegisterUser(NeolibBase):
    """Provides an interface for registering a new user on Neopets

    This class contains all of the functionality for walking through the
    Neopets registration process. It provides two public methods, one for
    setting up the information necessary for the registration and one for
    actually creating a new user. Please refer to the setup function prior to
    utilizing this class as the registration process looks for very specific
    information and the setup process can not provide a complete acurracy test
    to confirm that the information you submitted will be accepted.

    Attributes:
        | **username**: The username to register
        | **password**: The password to use for the new account
        | **country**: The country of the new user
        | **state**: The state of the new user
        | **gender**: The gender of the new user
        | **birth_day**: The day of the new user's birthday
        | **birth_month**: The month of the new user's birthday
        | **birth_year**: The year of the new user's birthday
        | *email_address**: The email address to register the user with

        | **neopet_name**: The name of the neopet to register
        | **neopet_species**: The species of the neopet to register
        | **neopet_color**: The color of the neopet to register
        | **neopet_gender**: The gender of the neopet to register
        | **neopet_terrain**: The type of terrain the new neopet likes (0-7)
        | **neopet_likes**: The likes of the new neopet (0-5)
        | **neopet_meetothers**: What the new neopet does when it meets others (0-7)
        | **neopet_stats**: The stats to select for the new neopet (1-3)

        | **step_delay**: The maximum delay time (in seconds) to wait between
            each registration step (min = max / 2)

        | **COLORS**: A list of acceptable colors for the new neopet
        | **TERRAINS**: A list of acceptable terrains for the new neopet
        | **LIKES**: A list of acceptable likes for the new neopet
        | **MEETOTHERS**: A list of acceptable options for meeting others
        | **STATS**: A list of acceptable stats for the new neopet
    """
    username = ''
    password = ''
    country = ''
    state = ''
    gender = ''
    birth_day = ''
    birth_month = ''
    birth_year = ''
    email_address = ''

    neopet_name = ''
    neopet_species = ''
    neopet_color = ''
    neopet_gender = ''
    neopet_terrain = ''
    neopet_likes = ''
    neopet_meetothers = ''
    neopet_stats = ''

    step_delay = 10

    COLORS = ['green', 'blue', 'red', 'yellow']

    TERRAINS = {
        'Forest': 0,
        'Sea': 1,
        'River': 2,
        'Jungle': 3,
        'Town': 4,
        'City': 5,
        'Mountains': 6,
        'Wildernes': 7,
    }

    LIKES = {
        'Gathering Food': 0,
        'Exploring The Land': 1,
        'Making Friends': 2,
        'Hunting For Treasure': 3,
        'Pestering Others': 4,
        'Reading and Learning': 5,
    }

    MEETOTHERS = {
        'Act Very Friendly': 0,
        'Smile Sweetly': 1,
        'Approach with Caution': 2,
        'Stand their Ground': 3,
        'Run Awaaaay!!!': 4,
        'Insult from Afar': 5,
        'Attack if they are Weaker': 6,
        'Try and make Friends': 7,
    }

    STATS = [1, 2, 3]

    _log_name = 'neolib.registration.RegisterUser'

    _urls = {
        'index': 'http://www.neopets.com/signup/index.phtml',
        'ajax': 'http://www.neopets.com/signup/ajax.phtml',
        'step': 'http://www.neopets.com/signup/index.phtml?cookieCheck=1',
        'neopet': 'http://www.neopets.com/reg/process_page6.phtml',
        'neopet_step': 'http://www.neopets.com/reg/page4.phtml',
        'neopet_check': '?neopet_name=%s&suggest_name=0&check_availability=1&r=908',

    }

    def __init__(self):
        """Initializes the class instance"""
        super().__init__()

    def setup(self, config):
        """Sets up the class instance for registering a new user

        Some internal checking for validity of the given configuration values
        is done and appropiate exceptions will be raised if any validation
        fails. However, not all data can be checked in this function. To
        alleviate problems registering, please refer to the example below for
        the format in which data should be passed to this function. Please note
        that this function will fail with an exception if the dictionary
        supplied is not formatted exactly as shown in the example blelow.

        Args:
            | **attributes**: A dictionary of configuration data to be used

        Example:
            ru = RegisterUser()
            config = {
                'general': {
                    'username': 'myusername',
                    'password': 'mypassword',
                    'country': 'US',
                    'state': 'WA',
                    'gender': 'M',
                    'birth_day': '01',
                    'birth_month': '10',
                    'birth_year': '1984',
                    'email_address': 'me@mysite.com',
                },
                'neopet': {
                    'name': 'myneopetname',
                    'species': 'aisha',
                    'color': 'green',
                    'gender': 'male',
                    'terrain': '2',
                    'likes': '1',
                    'meetothers': '6',
                    'stats': '2',
                },
            }

            ru.setup(config)
        """
        # Need to force the end-user to provide all attributes
        required_gen = ['username', 'password', 'country', 'state', 'gender',
                        'birth_day', 'birth_month', 'birth_year', 'email_address']
        required_neopet = ['name', 'species', 'color', 'gender', 'terrain',
                           'likes', 'meetothers', 'stats']

        for r in required_gen:
            if r not in config['general'].keys():
                raise MissingRequiredAttribute('Missing required general attribute ' + r)

        for r in required_neopet:
            if r not in config['neopet'].keys():
                raise MissingRequiredAttribute('Missing required neopet attribute ' + r)

        # We can check the validity of the Neopets information
        if config['neopet']['color'] not in self.COLORS:
            raise InvalidNeopet(config['neopet']['color'] + ' is not valid')

        if int(config['neopet']['terrain']) not in self.TERRAINS.values():
            raise InvalidNeopet(config['neopet']['terrain'] + ' is not valid')

        if int(config['neopet']['likes']) not in self.LIKES.values():
            raise InvalidNeopet(config['neopet']['likes'] + ' is not valid')

        if int(config['neopet']['meetothers']) not in self.MEETOTHERS.values():
            raise InvalidNeopet(config['neopet']['meetothers'] + ' is not valid')

        if int(config['neopet']['stats']) not in self.STATS:
            raise InvalidNeopet(config['neopet']['stats'] + ' is not valid')

        # Now we just try to set the config
        try:
            for key in config['general'].keys():
                setattr(self, key, config['general'][key])
        except Exception:
            raise AttributeNotFound('General attribute `' + key + '` not found')

        try:
            for key in config['neopet'].keys():
                setattr(self, 'neopet_' + key, config['neopet'][key])
        except Exception:
            raise AttributeNotFound('Neopet attribute `' + key + '` not found')

        # Setup the user instance
        self._usr = User(self.username, self.password)

    def register(self):
        """Registers a new user with the configured data and returns the result

        This function takes the configured data passed during setup and
        proceeds to step through all the steps of the registration process. The
        process is split into four sections and each section by default has a
        random delay inserted before moving on to the next section to provide a
        more realistic user registration process. This delay can be overriden
        by setting the `step_delay` attribute to 0. Each step is checked for a
        successful submittal and an exception will be raised if any step fails.
        If the registration is successful a newly configured :class:`.User`
        instance will be returned with the new user account already logged in.

        Returns:
            A :class:`.User` instance configured with the newly registered user
        """
        # Setup the cookies by requesting the sign-up page
        self._get_page('index')

        if not self._check_username():
            raise UsernameNotAvailable(self.username + ' is taken')

        # Step one, submit username and password
        j = self._step_one()

        if not j['success']:
            msg = self._to_element(j['message']).text_content()
            self._logger.error('Failed step 1 with result: ' + msg)
            raise InvalidPassword('`' + self.password + '` is not valid')

        # Need to ensure next step is requested to preserve cookies
        self._get_page('step')

        # Pause for realism
        self._wait_random(self.step_delay)

        # Step two, submit general user information
        j = self._step_two()

        if not j['success']:
            msg = self._to_element(j['message']).text_content()
            self._logger.error('Failed step 2 with result: ' + msg)
            raise InvalidDetails('Second step of registration failed')

        self._get_page('step')

        self._wait_random(self.step_delay)

        # Step three, submit the email address
        j = self._step_three()

        if not j['success']:
            msg = self._to_element(j['message']).text_content()
            self._logger.error('Failed step 3 with result: ' + msg)
            raise InvalidEmail('Invalid email address')

        self._get_page('neopet_step')

        self._wait_random(self.step_delay)

        if not self._check_neopet():
            raise NeopetNotAvailable('The neopet name `' + self.neopet_name +
                                     'is taken')

        # The final step is to create the Neopet
        result = self._create_neopet()

        if 'pet_success' not in result:
            self._logger.error('Failed to create neopet: ' + result)
            raise InvalidNeopet('Failed to create the Neopet')

        return self._usr

    def _step_one(self):
        """Performs step one of the registration process

        Step one of the registration process is submitting the username and
        password as well as accepting the EULA.

        Returns:
            The returned JSON from the form submission
        """
        data = {
            'method': 'step1',
            'username': self.username,
            'password1': self.password,
            'password2': self.password,
            'terms': 'true',
            'destination': ''
        }

        self._logger.debug('Initiating step 1 with: ' + str(data))
        pg = self._get_page('ajax', post_data=data)

        return pg.json

    def _step_two(self):
        """Performs step two of the registration process

        Step two of the registration process is submitting the personal
        information of the new user including birth date and geographical
        location.

        Returns:
            The returned JSON from the form submission
        """
        data = {
            'method': 'step2',
            'name': '',
            'city': '',
            'year': self.birth_year,
            'zip': '',
            'country': self.country,
            'month': self.birth_month,
            'gender': self.gender,
            'usState': self.state,
            'day': self.birth_day
            }

        self._logger.debug('Initiating step 2 with: ' + str(data))
        pg = self._get_page('ajax', post_data=data)

        return pg.json

    def _step_three(self):
        """Performs step three of the registration process

        Step three of the registration process is submitting the email address
        for the new user account

        Returns:
            The returned JSON from the form submission
        """
        data = {
            'method': 'step3',
            'email1': self.email_address,
            'email2': self.email_address,
            'optinNeopets': 'false',
            'optinEmail': ''
        }

        self._logger.debug('Initiating step 3 with: ' + str(data))
        pg = self._get_page('ajax', post_data=data)

        return pg.json

    def _create_neopet(self):
        """Creates a new Neopet with the configured data

        The last step of the registration process is creating the user's first
        Neopet. This function covers that last step.

        Returns:
            The returned HTML content from the form submission
        """
        data = {
            'neopet_name': self.neopet_name.lower(),
            'selected_pet': self.neopet_species,
            'selected_pet_colour': self.neopet_color,
            'gender': self.neopet_gender,
            'terrain': self.neopet_terrain,
            'likes': self.neopet_likes,
            'meetothers': self.neopet_meetothers,
            'pet_stats_set': self.neopet_stats,
        }

        self._logger.debug('Creating neopet with: ' + str(data))
        pg = self._get_page('neopet', post_data=data)

        return pg.content

    def _check_username(self):
        """Checks if a username is already taken or not

        Returns:
            Boolean value indicating if the username is available
        """
        data = {'method': 'checkAvailability', 'username': self.username}

        pg = self._get_page('ajax', post_data=data)

        return pg.json['success']

    def _check_neopet(self):
        """Checks if a neopet name is already taken or not

        Returns:
            Boolean value indicating if the neopet name is available
        """
        pg = self._usr.get_page(self._urls['neopet'] +
                                (self._urls['neopet_check'] % self.neopet_name))

        return 'is taken' not in pg.content

    def __repr__(self):
        return "User Registration <" + self.username + ">"



from collections import UserList

from neolib.Exceptions import ParseException
from neolib.NeolibBase import NeolibBase
from neolib.shop.Transaction import Transaction


class History(NeolibBase, UserList):
    """ Represents the sales history for a user's shop

    Attributes
        **data**: A list of :class:`Transaction` instances showing the history
    """
    data = []

    _log_name = 'neolib.shop.Transaction'

    _urls = {
        'sales': 'http://www.neopets.com/market.phtml?type=sales',
    }

    _paths = {
        'rows': '//*[@id="content"]/table/tr/td[2]/table[2]/tr',
        'details': './td'
    }

    def __init__(self, usr):
        super().__init__(usr)

    def load(self):
        """ Loads the user's sales history """
        # Load the history page
        pg = self._get_page('sales')

        try:
            rows = self._xpath('rows', pg)
            rows.pop(0)
            rows.pop()

            for row in rows:
                trans = Transaction()
                details = self._xpath('details', row)

                trans.date = details[0].text_content()
                trans.item = details[1].text_content()
                trans.buyer = details[2].text_content()
                trans.price = int(self._remove_multi(details[3].text_content(), [',', ' NP']))

                self.data.append(trans)
        except Exception:
            self._logger.exception('Failed to parse shop history', {'pg': pg})
            raise ParseException('Failed to parse shop history')

    def __repr__(self):
        return 'Sales History <' + str(len(self.data)) + ' Transactions>'



import logging
from datetime import datetime

#from neolib import LOG_DIR, PAGE_DIR

LOG_DIR = 'neolib_logs'
LOG_FILE = 'neolib.log'
PAGE_DIR = 'pages'


class HTMLFilter(logging.Filter):
    """Allows for proper filtering of HTML logging
    """

    def filter(self, record):
        """ Called when a log entry is made, ensures page content is logged properly

        Attributes
           record (Record) -- The log record

        Returns
           bool - If the log entry should be logged
        """
        # HTML needs special handling
        if 'pg' in record.args:
            self.handleHTML(record)

        return True

    def handleHTML(self, record):
        """ Saves the given record's page content to a .html file

        Attributes
           record (Record) -- The log record
        """
        # Create a unique file name to identify where this HTML source is coming from
        fileName = datetime.today().strftime("Neolib %Y-%m-%d %H-%M-%S ") + record.module + ".html"

        # Sometimes module may encase the text with < > which is an invalid character for a file name
        fileName = fileName.replace("<", "").replace(">", "")

        # Grab the pg
        pg = record.args['pg']

        # Format a log message that details the page
        ret = "Message: " + record.msg + "\nLine Number: " + str(record.lineno) + "\nURL: " + str(pg.url) + "\nPost Data: " + str(pg.post_data) + "\nAdditional Vars: " + str(pg.header_values)
        ret += "\n\n\n" + str(pg.response.request.headers) + "\n\n" + pg.content

        # Write the file
        f = open(LOG_DIR + '/' + PAGE_DIR + '/' + fileName, "w", encoding='utf-8')
        f.write(ret)
        f.close()



from collections import UserList

from neolib.Exceptions import ParseException
from neolib.item.ItemList import ItemList


class InventoryItemList(ItemList):
    """ Represents a returned list of items from querying a user's inventory """

    SDB = 'deposit'
    DONATE = 'donate'
    DROP = 'discard'
    SHOP = 'stock'
    GALLERY = 'gallery'

    _log_name = 'neolib.item.InventoryItemList'

    _urls = {
        'quickstock': 'http://www.neopets.com/quickstock.phtml'
    }

    def __init__(self, usr, items):
        super().__init__(usr, items)

    def move(self, location):
        """ Moves all items in the current list to the destination specified

        Arguments:
            | **location**: The destination to move the items to

        Returns:
            Boolean inicating if the move was successful

        Example:
            >>> snowballs = usr.inventory.find(name__contains='Snowball')
            >>> snowballs.move(snowballs.SHOP)
            True
        """
        pg = self._get_page('quickstock')
        form = pg.form(action='process_quickstock.phtml')[0]

        # First we need to build a dictionary of positions and id's
        ids = {}
        try:
            for name, inp in form.items():
                if 'id_arr' in name:
                    pos = inp.name.split('[')[1].replace(']', '')
                    ids[pos] = inp.value
        except Exception:
            self._logger.exception('Unable to parse item id\'s')
            raise ParseException('Unable to parse item id\'s')

        # Next we need to remove all positions that are not changing
        remove = []
        for pos in ids.keys():
            found = False
            for item in self.data:
                if item.id == ids[pos]:
                    found = True

            if not found:
                remove.append(pos)

        for pos in remove:
            del ids[pos]

        # Now we need to remove all radio fields that are not changing
        try:
            for name, inp in list(form.items()):
                if 'radio_arr' in name:
                    pos = inp.name.split('[')[1].replace(']', '')
                    if pos not in list(ids.keys()):
                        del form[name]
        except Exception:
            self._logger.exception('Unable to parse radio fields')
            raise ParseException('Unable to parse radio fields')

        # Finally we need to set the destination for remaining radio fields
        if location == self.SDB:
            value = self.SDB
        elif location == self.DONATE:
            value = self.DONATE
        elif location == self.DROP:
            value = self.DROP
        elif location == self.SHOP:
            value = self.SHOP
        elif location == self.GALLERY:
            value = self.GALLERY
        else:
            return False

        for name, inp in list(form.items()):
            if 'radio_arr' in name:
                inp.value = value

        # Now we submit the form and check the result
        pg = form.submit(self._usr)
        if 'red_oops.gif' in pg.content:
            return False
        else:
            return True

    def __getitem__(self, item):
        result = UserList.__getitem__(self, item)
        if type(result) is list:
            if len(result) > 1:
                return InventoryItemList(self._usr, result)
        else:
            return result



from neolib.item.Item import Item


class USFrontItem(Item):
    """ Represents an item in another user's shop page

    Attributes:
        | **url**: The url to purchase this item
        | **stock**: The amount of this item stocked in the user's shop
    """
    url = ''
    stock = 0

    _log_name = 'neolib.item.USFrontItem'

    def buy(self):
        """ Attempts to buy this item from the user's shop

        Returns:
            Boolean value indicating if the purchase was successful
        """
        # Attempt to buy the item
        pg = self._usr.get_page(self._base_url + '/' + self.url)

        if 'does not exist in this shop' in pg.content:
            return False
        else:
            return True

    def __repr__(self):
        return 'User Shop Front Item <' + self.name + '>'



from neolib.item.Item import Item


class SDBItem(Item):
    """ Represents an item in a user's safety deposit box """
    stock = 0
    pg = 0

    remove = False

    _log_name = 'neolib.item.SDBItem'

    def __repr__(self):
        return "SDB Item <" + self.name + ">"



from neolib.Exceptions import ParseException
from neolib.inventory.Inventory import Inventory
from neolib.item.USFrontItem import USFrontItem
from neolib.item.USFrontItemList import USFrontItemList


class USFrontInventory(Inventory):
    """ Provides an interface to another user's shop inventory """

    _log_name = 'neolib.inventory.USFrontInventory'

    _urls = {
        'shop': 'http://www.neopets.com/browseshop.phtml?lower=%s&owner=%s'
    }

    _paths = {
        'rows': '//*[@id="content"]/table/tr/td[2]/table/tr',
        'tds': './td',
    }

    def load(self, owner, index, pages=False):
        """ Loads the shop contents

        Arguments:
            | **owner**: The owner of the shop
            | **index**: The index page of the shop
            | **pages**: Whether or not to fetch all shop pages
        """
        # Determine if we're going to load multiple pages
        try:
            if pages:
                # Parse the index
                self._parse_page(index)

                # Keep going until we run out of pages
                i = 1
                while True:
                    lim = i * 80

                    pg = self._get_page('shop', (str(lim), owner))
                    self._parse_page(pg)

                    if 'Next 80 Items' not in pg.content:
                        break

                    i += 1
            else:
                self._parse_page(index)
        except Exception:
            self._logger.exception('Failed to parse user front shop with owner: ' + owner, {'pg': pg})
            raise ParseException('Failed to parse user front shop')

    def find(self, **kwargs):
        """ Overrides the :class:`Inventory`:`find()` function to return an
        instance of :class:`USFrontItemList`.

        See the base class's function for more details
        """
        result = super().find(**kwargs)
        return USFrontItemList(self._usr, result)

    def _parse_page(self, pg):
        # Loop through rows of items
        for row in self._xpath('rows', pg):
            # Each td is an item
            for td in self._xpath('tds', row):
                item = USFrontItem('', self._usr)

                item.url = td.xpath('./a/@href')[0]
                item.name = str(td.xpath('./b/text()')[0])
                item.stock = int(td.xpath('./text()[3]')[0].replace(' in stock', ''))
                item.price = int(self._remove_multi(td.xpath('./text()[4]')[0], [',', ' NP', 'Cost : ']))

                self.data.append(item)

    def __repr__(self):
        return "User Front Shop Inventory <" + str(len(self.data)) + ">"



import logging
import random
import re
import time
from functools import reduce

from lxml import etree, html

from neolib.http.Page import Page


class NeolibBase:
    """The base class for most other classes in the Neolib library

    This class should be inherited by any class that is doing work on a
    :class:`User` object in the library. It provides the established template
    for doing such work and the original template should be followed as closely
    as possible to maintain uniformity throughout the library.

    Attributes:
        | **_log_name**: The name of the log being used by the child class
        | **_logger**: A logger instance used to write to the error log

        | **_usr**: A :class:`.User` instance being used by the child class for work

        | **_urls**: A dictionary of URLs that will be used for doing work
        | **_paths**: A dictionary of XPath queries that will be used for doing
            work. This dictionary should either be in {key: expression} format
            or {key {key: expression}} format
        | **_regex**: A dictionary for regular expressions that will be used for
            doing work. This dictionary should either be in {key: expression}
            format or {key: {key: expression}} format

        | **_base_url**: The base Neopets URL
    """

    _log_name = 'neolib'
    _logger = None

    _usr = None

    _urls = {}
    _paths = {}
    _regex = {}

    _base_url = 'http://www.neopets.com'

    def __init__(self, usr=None):
        """Initializes the base class by initializing the logger instance"""
        self._logger = logging.getLogger(self._log_name)
        self._usr = usr

    def _get_page(self, url, args=None, post_data='', header_values=''):
        """ Returns a :class:`Page` object representing the given url

        Args:
            | **url**: The name of the url to use for querying as stored
            in _urls. Should be in the format of 'key1/key2/key3'.

        Returns:
            :class:`Page` instance representing the given url
        """
        url = reduce(dict.__getitem__, url.split('/'), self._urls)

        if args:
            url = url % args

        return self._usr.get_page(url, post_data, header_values)

    def _search(self, exp, subject, all=False):
        """Searches the given string using the given expression name

        Args:
            | **exp**: The name of the expression to use for querying as stored
                in _regex. Should be in the format of 'key1/key2/key3'.
            | **subject*: Instance of :class:`.Page`, :class:`HTMLElement`, or
                a string value to search
            | **all**: Whether to search with DOTALL

        Returns:
            A list of matches from the regular expression query
        """
        if type(subject) is Page:
            string = subject.content
        elif type(subject) is html.HtmlElement:
            string = self._to_html(subject)
        else:
            string = subject

        query = reduce(dict.__getitem__, exp.split('/'), self._regex)

        if all:
            return re.findall(query, string, re.DOTALL)
        else:
            return re.findall(query, string)

    def _xpath(self, path, subject):
        """Searches the given subject using the given xpath name

        Args:
            | **path**: The name of the path to use for querying as stored in
                _paths. Should be in the format of 'key1/key2/key3'.
            | **subject*: Instance of :class:`.Page`, or :class:`HTMLElement`

        Returns:
            A list of matches from the xpath query
        """
        if type(subject) is Page:
            ele = subject.document
        else:
            ele = subject

        query = reduce(dict.__getitem__, path.split('/'), self._paths)

        return ele.xpath(query)

    def _xpath_page(self, url, path):
        """ Fetches the given url and applies the given xpath to the page
        Args:
            | **url**: The name of the url to use for querying as stored in
                _urls. Should be in the format of 'key1/key2/key3'.
            | **path**: The name of the xpath to use for querying as stored in
                _paths. Should be in the format of 'key1/key2/key3'.

        Returns:
            A list of matches from the xpath query
        """
        return self._xpath(path, self._get_page(url))

    def _to_element(self, string):
        """Converts a HTML string into a lxml element

        Args:
            **string**: The HTML string to convert
        """
        return html.document_fromstring(string)

    def _to_html(self, element):
        """Converts a html element to a string

        Args:
            **element**: The HTML element to conver to a HTML string
        """
        return etree.tostring(element).decode('utf-8')

    def _path_to_html(self, path, subject):
        """Converts an xpath query to HTML

        Note that this function will fail if the xpath query returns no
        results. Furthermore, if the query returns more than one result only
        the first result will be converted to HTML and returned.

        Args:
            **path**: The name of the expression to use for querying as stored
            in _paths. Should be in the format of 'key1/key2/key3'.
            **subject**: Instance of :class:`.Page`, or :class:`HTMLElement`
        """
        return self._to_html(self._xpath(path, subject)[0])

    def _wait_random(max):
        """Wait a random time up to the max seconds given

        Takes the max seconds given and divides it by two to obtain a minimum
        wait value. Picks a random number between the max and minimum value to
        wait.

        Args:
            **max**: The max number of seconds to wait
        """
        if max == 0:
            return

        min = max / 2
        delay = random.randint(min, max)

        time.sleep(delay)

    def _is_init(self, string):
        """Determines whether a string can be safely casted to an integer

        Args:
            **string**: The string to test

        Returns
            Boolean result representing if the cast was successful or not
        """
        try:
            int(string)
            return True
        except Exception:
            return False

    def _remove_multi(self, string, substrings):
        """ Removes all instances of given substrings from the given string

        Args:
            | **string**: The string to replace in
            | **substrings**: The substrings to remove from the given string

        Returns:
            Modified string with substrings removed
        """
        for sub in substrings:
            string = string.replace(sub, '')

        return string

    def _format_nps(self, nps_str):
        """ Formats common neopoints strings into integer values

        Arguments:
            nps_str: The string to format (i.e '21,000 NP')

        Returns:
            An integer value reflecting the string value
        """
        return int(self._remove_multi(nps_str, [',', 'NP', ' ', 'np']))



from neolib.Exceptions import ParseException
from neolib.NeolibBase import NeolibBase
from neolib.user.Neopet import Neopet


class Profile(NeolibBase):
    """Represents the Neopets profile page of a user

    This class holds most (not all) of the information found on the profile
    page of a user. This includes the user's general information, statistics
    about their collections, shop and gallery information, and general
    information about all of their neopets.

    Attributes
       | **name**: The real name of the user (often not given)
       | **age**: The age of the user's accounts in months
       | **gender**: The gender of the user
       | **country**: The user's country of origin
       | **last_spotted**: A string describing the last time the user was seen
       | **started_playing**: The date the user started playing Neopets
       | **hobbies**: Any hobbies the user has (often not given)

       | **secret_avatars**: The number of secret avatars the user has
       | **keyquest_tokens**: The number of Keyquest tokens the user has
       | **stamps**: The number of stamps the user has
       | **neocards**: The total number of neocards the user has
       | **site_themes**: The total number of site themes the user has
       | **bd_wins**: The total number of one-player battledome wins the user has

       | **neopets**: A list of :class:`.Neopet` objects representing the user's pets

       | **shop_name**: The name of the user's shop
       | **shop_size**: The size of the user's shop
       | **shop_link**: The partial url to the user's shop

       | **gallery_name**: The name of the user's gallery
       | **gallery_size**: The size of the user's gallery
       | **gallery_link**: The partial url to the user's gallery
    """

    name = ''
    age = 0
    gender = ''
    country = ''
    last_spotted = ''
    started_playing = ''
    hobbies = ''

    secret_avatars = 0
    keyquest_tokens = 0
    stamps = 0
    neocards = 0
    site_themes = 0
    bd_wins = 0

    neopets = []

    shop_name = ''
    shop_size = 0
    shop_link = ''

    gallery_name = ''
    gallery_size = 0
    gallery_link = ''

    _log_name = 'neolib.user.profile'

    _urls = {
        'profile': 'http://www.neopets.com/userlookup.phtml?user=%s'
        }

    _paths = {
        'general': '//*[@id="userinfo"]/table/tr[2]/td/table/tr[1]/td',
        'colls1': '//*[@id="usercollections"]/table/tr[2]/td/table/tr/td[1]',
        'colls2': '//*[@id="usercollections"]/table/tr[2]/td/table/tr/td[2]',
        'shop_gallery': '//*[@id="usershop"]/table/tr[2]/td',
        'neopets': '//*[@id="userneopets"]/table/tr[2]/td/table/tr/td',
        }

    _regex = {
        'general': {
            'age': 'shields/(.*?).gif',
            'name': 'Name:</b>(.*?)<br/>',
            'gender': 'Gender:</b> <b.*">(.*?)</b>',
            'country': 'Country:</b>(.*?)<br/>',
            'last_spotted': 'Last Spotted:</b>(.*?)<br/>',
            'started_playing': 'Started Playing:</b>(.*?)<br/>',
            'hobbies': 'Hobbies:</b>(.*?)<br/>',
            },
        'colls1': {
            'secret_avatars': 'Secret Avatars:</b><br/>(.*?)<br/>',
            'stamps': 'Stamps:</b><br/>(.*?)<br/>',
            'site_themes': 'Site Themes:</b><br/>(.*?)</td>',
            },
        'colls2': {
            'keyquest_tokens': 'Key Quest Tokens:</b>.*?<br/>(.*?) \(',
            'neocards': 'Neodeck:.*?<br/>(.*?) cards',
            'neocards2': '</a><br/><b>(.*?)</b> cards',
            'bd_wins': '"><b>(.*?) out of',
            },
        'shop_gallery': {
            'shop_name': '<a href=".*?"><b>(.*?)</b>',
            'shop_size': 'Size:</b>(.*?)<br/>',
            'shop_link': '<a href="(.*?)"><b>',
            'gallery_name': 'Gallery.*?">(.*?)</a>',
            'gallery_size': 'Gallery.*?</b> ([1-9]+).*</td>',
            'gallery_link': 'Gallery:</b> <a href="(.*?)">',
            },
        'neopets': {
            'name': '<b>(.*?)</b><br/>',
            'gender': ';">(.*?)</b>',
            'species': '</b> (.*?)<br/>',
            'age': 'Age:</b> (.*?) days',
            'level': 'Level:</b> (.*?)<',
            },
        }

    def __init__(self, usr):
        """Initializes the profile with the given :class:`User` object

        Args:
            **user**: The :class:`User` object to load the profile for
        """
        super().__init__(usr)

    def load(self):
        """Fetches the user profile data and populates the attributes

        Query's the user's profile on Neopets using the username of the given
        :class:`User` object when the profile was initialized. This method can
        be called multiple times to update the loaded profile details.

        Raises:
            **ParseException**: An error occured parsing the user's profile
        """
        # Much of the user profile can change without much uniformity. This
        # means certain HTML elements may or may not be in different places.
        # It all depends on how much the user has filled their profile in.
        # Therefore, this class uses a combination of xpath and regular
        # expressions to parse all required information from the profile.

        # Since the pattern for parsing with regular expressions and xpath is
        # fairly similar across all the sections of a user profile, a
        # refactored function was implemented which takes an exception function
        # for dealing with any exceptional fields when running through a
        # profile section.

        # The age of an account is only displayed in the form of an image. This
        # image can represent either the number of weeks, months, or years. The
        # below exception function determines which type of image it is and
        # calculates the proper age.
        def general_exception(self, key, result, html):
            if key == 'age':
                if 'years' in result:
                    self.age = int(result.split("_")[0]) * 12
                elif 'mth' in result:
                    self.age = int(result.split('mth')[0])
                else:
                    self.age = int(result.split('wk')[0]) / 4
                return True
            else:
                return False

        # The number of neocards a user has is display differently depending on
        # if the user has an active deck. The below exception function
        # determines which case exists and parses accordingly.
        def collections2_exception(self, key, result, html):
            if key == 'neocards':
                # If the first match failed it will match an entire sentence.
                # So we just check the results length to see if it failed
                if len(result) > 4:
                    result = self._search('colls2/neocards2', html, True)[0]

                    self.neocards = int(self._remove_extra(result))
                else:
                    self.neocards = int(self._remove_extra(result))
                return True
            elif key == 'neocards2':
                return True
            else:
                return False

        # Get the profile page
        pg = self._get_page('profile', self._usr.username)

        try:
            # Parse the general profile details
            self._set_attributes(pg, 'general', 'general', general_exception)

            # Parse the first set of collections
            self._set_attributes(pg, 'colls1', 'colls1')

            # Parse the second set of collections
            self._set_attributes(pg, 'colls2', 'colls2', collections2_exception)

            # The user might not have a shop or gallery
            if 'Shop & Gallery' in pg.content:
                self._set_attributes(pg, 'shop_gallery', 'shop_gallery')

            # The neopets are parsed slightly differently
            for td in pg.xpath(self._paths['neopets']):
                html = self._to_html(td)
                pet = Neopet()

                for key in self._regex['neopets'].keys():
                    result = self._search('neopets/' + key, html)
                    if result:
                        setattr(pet, key, self._remove_extra(result[0]))
                self.neopets.append(pet)
        except:
            self._logger.exception('Failed to parse user profile with name: ' + self._usr.username, {'pg': pg})
            raise ParseException('Could not parse user profile')

    def _set_attributes(self, pg, path, exps, exception=None):
        html = self._path_to_html(path, pg)

        # Loop through all supplied regular expressions
        for exp in self._regex[exps].keys():
            # Search the supplied HTML for matches
            result = self._search(exps + '/' + exp, html)

            # Sometimes weird characters mess with the match so we try it one
            # more time with DOTALL
            if not result:
                result = self._search(exps + '/' + exp, html, True)

            if result:
                # If an exception function was given, call it and continue if
                # it passes
                if exception:
                    if exception(self, exp, result[0], html):
                        continue
                # Set this classes attribute with the cleaned up match
                setattr(self, exp, self._remove_extra(result[0]))

    def _remove_extra(self, string):
        # Remove all the extra ugly from any matches
        clean_string = string.strip()
        clean_string = clean_string.replace('\n', '')
        clean_string = clean_string.replace('\r', '')
        clean_string = clean_string.replace('\t', '')

        return clean_string

    def __repr__(self):
        return "Profile <" + self._usr.username + ">"



from neolib.Exceptions import InvalidItemID, ParseException
from neolib.item.Item import Item


class InventoryItem(Item):
    """ Represents an item in the user's current inventory

    Attributes:
        | **weight**: The item weight
        | **rarity**: The item rarity
        | **value**: The item value
    """
    weight = ''
    rarity = ''
    value = ''

    SDB = 'safetydeposit'
    DONATE = 'donate'
    DROP = 'drop'
    SHOP = 'stockshop'
    GALLERY = 'stockgallery'
    # GIVE = 'give' (FUTURE)
    # AUCTION = 'auction' (FUTURE)

    _log_name = 'neolib.item.InventoryItem'

    _urls = {
        'item': 'http://www.neopets.com/iteminfo.phtml?obj_id=%s'
    }

    _paths = {
        'image': '/html/body/table[1]/tr/td[1]/img/@src',
        'name': '/html/body/table[1]/tr/td[2]/text()[2]',
        'desc': '/html/body/div[2]/span/i/text()',
        'type': '/html/body/table[2]/tr[1]/td[2]/text()',
        'weight': '/html/body/table[2]/tr[2]/td[2]/text()',
        'rarity': '/html/body/table[2]/tr[3]/td[2]/text()',
        'value': '/html/body/table[2]/tr[4]/td[2]/text()'
    }

    def __init__(self, id, usr, name=""):
        """ Initializes parent class with the item id and :class:`User` instance"""
        super().__init__(id, usr, name)

    def get_details(self):
        """ Fills in the item details obtained off the item info page

        Most details can be obtained directly off of the user's inventory
        screen, however a few key details like weight and value cannot and this
        function should be called to retrieve these """
        pg = self._get_page('item', self.id)

        if "not in your inventory" in pg.content:
            self._logger.error('No item with id ' + self.id + ' in inventory')
            raise InvalidItemID(self.id)

        try:
            self.name = str(self._xpath('name', pg)[0].replace(' : ', ''))
            self.img = str(self._xpath('image', pg)[0])
            self.desc = str(self._xpath('desc', pg)[0])
            self.type = str(self._xpath('type', pg)[0])
            self.weight = str(self._xpath('weight', pg)[0])
            self.rarity = str(self._xpath('rarity', pg)[0])
            self.value = str(self._xpath('value', pg)[0])
        except Exception:
            self._logger.exception('Failed to parse details for ' + self.name, {'pg': pg})
            raise ParseException('Failed to parse details for ' + self.name)

    def move(self, location):
        """ Moves an item from the user's inventory to the given location

        Args
            | **location**: The location to move the items

        Returns
            Boolean value indicating whether the move was successful or not

        Example:
            >>> item = usr.inventory.items[0]
            >>> item.move(item.SHOP)
            True
        """
        pg = self._usr.get_page(self._urls['item'] % self.id)

        form = pg.form(action='useobject.phtml')[0]

        if location == self.SDB:
            form.update(action=self.SDB)
        elif location == self.DONATE:
            form.update(action=self.DONATE)
        elif location == self.DROP:
            form.update(action=self.DROP)
        elif location == self.SHOP:
            form.update(action=self.SHOP)
        elif location == self.GALLERY:
            form.update(action=self.GALLERY)
        else:
            return False

        pg = form.submit(self._usr)
        if 'red_oops.gif' in pg.content:
            return False
        else:
            return True



from collections import UserList

from neolib.item.ItemList import ItemList


class WizardItemList(ItemList):
    """ Represents the results from a Shop Wizard search """

    query_types = [
        'contains',
        'startswith',
        'endswith'
        'gt',
        'lt',
    ]

    _log_name = 'neolib.item.WizardItemList'

    def buy(self):
        """ Attempts to buy all items in the current item list

        Returns:
            List of items that were purchased
        """
        success = []
        for item in self.data:
            if item.buy():
                success.append(item)

        return success

    def find(self, **kwargs):
        """ Searches the current results using keyword arguments

        Loops through each keyword argument and searches the current results
        for any item that has an attribute (key name) that matches a value
        (key value).

        Args
            **kwargs**: Keyword arguments used for searching

        Returns
            A list of items that match the search

        Example
            low = results.find(price__lt=1000)[0]
        """
        matches = []

        for item in self.data:
            match = 0
            for key in kwargs.keys():
                try:
                    # Determine if this query has extra arguments
                    if '__' in key:
                        arg = key.split('__')[1]
                        atrib = key.split('__')[0]
                        value = getattr(item, atrib)

                        if arg not in self.query_types:
                            continue

                        if arg == 'contains' and type(value) is str:
                            if kwargs[key] in value:
                                match += 1
                        elif arg == 'startswith' and type(value) is str:
                            if value.startswith(kwargs[key]):
                                match += 1
                        elif arg == 'endswith' and type(value) is str:
                            if value.endswith(kwargs[key]):
                                match += 1
                        elif arg == 'gt' and self._is_int(value):
                            if int(value) > int(kwargs[key]):
                                match += 1
                        elif arg == 'lt' and self._is_init(value):
                            if int(value) < int(kwargs[key]):
                                match += 1
                    else:
                        value = getattr(item, key)
                        if kwargs[key] == value:
                            match += 1
                except Exception:
                    continue

            if match == len(kwargs.keys()):
                matches.append(item)

        return WizardItemList(self._usr, matches)

    def __getitem__(self, item):
        result = UserList.__getitem__(self, item)
        if type(result) is list:
            if len(result) > 1:
                return WizardItemList(self._usr, result)
        else:
            return result

    def __repr__(self):
        return 'Shop Wizard Results <' + str(len(self.data)) + ' Items>'




# General
class ParseException(Exception):
    pass

# Item
class InvalidItemID(Exception):
    pass

# Registration
class MissingRequiredAttribute(Exception):
    pass

class AttributeNotFound(Exception):
    pass

class UsernameNotAvailable(Exception):
    pass

class InvalidPassword(Exception):
    pass

class InvalidDetails(Exception):
    pass

class InvalidEmail(Exception):
    pass

class NeopetNotAvailable(Exception):
    pass

class InvalidNeopet(Exception):
    pass

#Shop
class WizardBanned(Exception):
    pass

# User
class NeopetsOffline(Exception):
    pass

class BirthdayLocked(Exception):
    pass

class InvalidBirthday(Exception):
    pass

class AccountFrozen(Exception):
    pass

class NoActiveNeopet(Exception):
    pass



import io

from PIL import Image

from neolib.Exceptions import ParseException
from neolib.item.Item import Item


class MSItem(Item):
    """ Represents an item in a main shop

    Attributes:
        | **brr**: Something TNT uses for tracking main shop purchases
        | **stock**: The number of this item in the shop
        | **stock_id**: Unique ID for the item in the shop
    """
    brr = ''
    stock = 0
    stock_id = ''

    _log_name = 'neolib.item.MSItem'

    _urls = {
        'haggle': 'http://www.neopets.com/haggle.phtml?obj_info_id=%s&stock_id=%s&brr=%s',
    }

    _paths = {
        'captcha': '//form[@name="haggleform"]//input[@type="image"]/@src',
    }

    def buy(self, price=0):
        """ Attempts to buy the item from the main shop

        This function will take the given price (or use the price of the item
        from the shop if no price is given) and attempt to purchase the item. It
        automatically scans the generated OCR and attempts to find the x,y
        coordinate of the darkest pixel. Returns the result.

        Arguments:
            **price**: Optional price to buy the item at
        Returns:
            Boolean indicating if the purchase was successful
        """
        # Goto the haggle page
        pg = self._get_page('haggle', (self.id, self.stock_id, self.brr))

        # Did we get a price?
        if not price:
            price = self.price

        try:
            # Check to see if the item is still for sale
            if pg.form(action='haggle.phtml') and 'SOLD OUT' not in pg.content:
                form = pg.form(action='haggle.phtml')[0]

                # Download the image
                url = self._base_url + self._xpath('captcha', pg)[0]
                pg = self._usr.get_page(url)

                x, y = self._crack_OCR(io.BytesIO(pg.content))
                form.update(current_offer=str(price), x=str(x), y=str(y))

                # Attempt to buy the item
                pg = form.submit(self._usr)

                if "I accept" in pg.content:
                    return True
                else:
                    return False
            else:
                return False
        except Exception:
            self._logger.exception('Failed to handle haggle page', {'pg': pg})
            raise ParseException('Failed to handle haggle page')

    def _crack_OCR(self, img):
        # Open the image and convert to grayscale
        im = Image.open(img)
        im = im.convert('L')

        # Find the darkest point and create a box around it
        lo, hi = im.getextrema()
        lo = im.point(lambda x: x == lo)
        x, y, _, _ = lo.getbbox()

        return x, y

    def __repr__(self):
        return 'Main Shop Item <' + self.name + '>'



from distutils.core import setup

setup(
    name="neolib2",
    packages=["neolib",
              "neolib.http",
              "neolib.inventory",
              "neolib.item",
              "neolib.registration",
              "neolib.shop",
              "neolib.user",
              "neolib.user.hooks"
              ],
    version="0.1.7",
    description="Neopets automation library for Python",
    author="Joshua Gilman",
    author_email="joshuagilman@gmail.com",
    url="https://github.com/jmgilman/neolib2",
    download_url="https://github.com/jmgilman/neolib2/tarball/0.1",
    install_requires=[
        "requests==2.4",
        "lxml==3.4.1",
        "pillow==2.6",
        ],
    package_data={'neolib2': ['tests/*', 'examples/*']},
    keywords=["Neopets", "automation", "library"],
    classifiers=[],
    long_description="",
    )



from neolib.Exceptions import ParseException
from neolib.NeolibBase import NeolibBase


class Bank(NeolibBase):
    """ Provides an interface to a user's bank

    This class provides a general interface for interacting with a user's bank
    including getting bank account information, making withdrawals and deposits,
    and collecting interest.

    Attributes:
        | **type**: The type of the bank account
        | **balance**: The current account balance
        | **interest_rate**: The account's interest rate
        | **yearly_interest**: The account's yearly interest amount
        | **daily_interest**: The account's daily interest amount
    """
    type = ''
    balance = 0
    interest_rate = 0.0
    yearly_interest = 0
    daily_interest = 0
    interest_collected = False

    _log_name = 'neolib.user.bank'

    _urls = {
        'index': 'http://www.neopets.com/bank.phtml',
        'process': 'http://www.neopets.com/process_bank.phtml',
    }

    _paths = {
        'rows': '//*[@id="content"]/table/tr/td[2]/div[2]/table/tr[2]/td/table/tr',
        'daily': '//*[@id="content"]/table/tr/td[2]/table[2]/tr/td/div/table/tr[2]/td/b/text()',
    }

    def load(self):
        """ Loads all bank account details """
        # Get the index
        pg = self._get_page('index')

        # Check if the user has a bank account
        if 'I see you don\'t currently have an account with us' in pg.content:
            self._logger.warning('User ' + self._usr.username + ' does not have a bank account')
            return

        # Load the details
        try:
            rows = self._xpath('rows', pg)
            self.type = str(rows[0].xpath('./td[2]/text()')[0])
            self.balance = self._format_nps(rows[1].xpath('./td[2]/text()')[0])
            self.interest_rate = float(rows[2].xpath('./td[2]/b/text()')[0].replace('%', ''))
            self.yearly_interest = self._format_nps(rows[3].xpath('./td[2]/text()')[0])

            # Some user's don't have enough for daily interest
            if 'you might want to deposit a few more Neopoints' not in pg.content:
                self.daily_interest = self._format_nps(self._xpath('daily', pg)[0])
                self.interest_collected = True

            if 'You have already collected your interest today' in pg.content:
                self.interest_collected = True
        except Exception:
            self._logger.exception('Failed to parse user bank details', {'pg': pg})
            raise ParseException('Failed to parse user bank details')

    def withdraw(self, amount):
        """ Withdraws neopoints from the bank

        Arguments:
            **amount**: The amount to withdraw

        Returns:
            Boolean value indicating if the action was successful
        """
        return self._action('withdraw', amount)

    def deposit(self, amount):
        """ Deposits neopoints from the bank

        Arguments:
            **amount**: The amount to deposit

        Returns:
            Boolean value indicating if the action was successful
        """
        return self._action('deposit', amount)

    def collect_interest(self):
        """ Collects the user's daily interest

        Returns:
            Boolean value indicating if the action was successful
        """
        if self._action('interest'):
            self.interest_collected = True
            return True
        else:
            return False

    def _action(self, type, amount=0):
        # I would use a form here, but Neopets HTML is so jacked up here it's
        # easier to just POST the data
        data = {
            'type': type,
            'amount': str(amount),
        }

        # Submit the form
        pg = self._get_page('process', post_data=data)

        # Check the result
        if 'red_oops.gif' in pg.content:
            return False
        else:
            return True

    def __repr__(self):
        return "Bank <" + "{:,}".format(self.balance) + " NP>"



from neolib.NeolibBase import NeolibBase


class Hook(NeolibBase):
    """A base class for creating hooks on :class:`User` get_page() method

    This class lays out the template for creating hooks that will be executed
    after each page request done with the :class:`User` class. It has one
    single function that is called. An instance of the current :class:`User`
    instance will be passed along with the page that was requested.
    """

    def __init__(self):
        super().__init__()

    def execute(self, usr, pg):
        """ Called after a page is requested

        Args
            usr: The :class:`User` instance which requested the page
            pg: The :class:`Page` instance representing the returned page
        """
        pass



from collections import UserList

from neolib.item.ItemList import ItemList


class SDBItemList(ItemList):
    """ Represents a list of items returned from searching a user's SDB """

    def __getitem__(self, item):
        result = UserList.__getitem__(self, item)
        if type(result) is list:
            if len(result) > 1:
                return SDBItemList(self._usr, result)
        else:
            return result



from collections import UserList

from neolib.item.ItemList import ItemList


class MSItemList(ItemList):
    """ Represents a list of items returned from querying a main shop """

    _log_name = 'neolib.item.MSItemList'

    def buy(self):
        """ Attempts to buy all items in this list with their asking prices

        Returns:
            List of items that were successfully bought
        """
        successful = []
        for item in self.data:
            if item.buy():
                successful.append(item)

        return successful

    def __getitem__(self, item):
        result = UserList.__getitem__(self, item)
        if type(result) is list:
            if len(result) > 1:
                return MSItemList(self._usr, result)
        else:
            return result



from neolib.Exceptions import ParseException
from neolib.inventory.Inventory import Inventory
from neolib.item.MSItem import MSItem
from neolib.item.MSItemList import MSItemList


class MSInventory(Inventory):
    """ Represents the inventory of a main shop """

    _log_name = 'neolib.inventory.MSInventory'

    _urls = {
        'shop': 'http://www.neopets.com/objects.phtml?type=shop&obj_type=%s',
    }

    _paths = {
        'rows': '//*[@id="content"]/table/tr/td[2]/form[2]/div/table/tr/td/div/table/tr[2]/td/table/tr',
        'tds': './td',
    }

    def load(self, id, index=None):
        """ Loads the main shop inventory

        Arguments:
            | **id**: The id of the shop to load
            | **index**: Optional index page of the shop to load from
        """
        # Check if we already have the index page
        if not index:
            pg = self._get_page('shop', str(id))
        else:
            pg = index

        # Clear any existing data
        self.data = []

        try:
            for row in self._xpath('rows', pg):
                for td in self._xpath('tds', row):
                    url = td.xpath('./a/@href')[0]
                    onclick = td.xpath('./a/@onclick')[0]

                    id = url.split('obj_info_id=')[1].split('&')[0]
                    stock_id = url.split('stock_id=')[1].split('&')[0]
                    brr = onclick.split('brr=')[1].split('\'')[0]

                    item = MSItem(id, self._usr)
                    item.name = str(td.xpath('./b/text()')[0])
                    item.stock = int(td.xpath('./text()')[0].replace(' in stock', ''))
                    item.price = int(self._remove_multi(td.xpath('./text()')[1], ['Cost: ', ',', ' NP']))
                    item.stock_id = stock_id
                    item.brr = brr

                    self.data.append(item)
        except Exception:
            self._logger.exception('Failed to parse main shop inventory for id: ' + str(id), {'pg': pg})
            raise ParseException('Failed to parse main shop inventory')

    def find(self, **kwargs):
        """ Overrides the :class:`Inventory`:`find()` function to return an
        instance of :class:`MSItemList`.

        See the base class's function for more details
        """
        result = super().find(**kwargs)
        return MSItemList(self._usr, result)

    def __repr__(self):
        return 'Main Shop Inventory <' + str(len(self.data)) + ' Items>'



from neolib.inventory.MSInventory import MSInventory
from neolib.NeolibBase import NeolibBase


class MainShop(NeolibBase):
    """ Represents a main shop

    Provides an interface for interacting with a main shop including searching
    the inventory and buying items.

    Attributes:
        **inventory**: Instance of :class:`MSInventory` with the shop's items
    """
    inventory = None

    _log_name = 'neolib.shop.MainShop'

    _urls = {
        'shop': 'http://www.neopets.com/objects.phtml?type=shop&obj_type=%s',
    }

    _paths = {
        'name': '//*[@id="content"]/table/tr/td[2]/table[1]/tr/td[1]/div/table/tr[1]/td/text()',
    }

    def __init__(self, usr, id):
        """ Loads the shop's name and inventory

        Arguments:
            | **usr**: The :class:`User` instance to use
            | **id**: The id of the shop to load
        """
        super().__init__(usr)

        # Request the shop page
        pg = self._get_page('shop', str(id))

        # Get the shop name
        self.name = str(self._xpath('name', pg)[0]).strip()

        # Check for items and load inventory
        if 'Sorry, we are sold out of everything!' not in pg.content:
            self.inventory = MSInventory(self._usr)
            self.inventory.load(id, pg)

    def __repr__(self):
        return 'Main Shop <' + self.name + '>'
